function [ kalman_targets ] = kalman_filtering( kalman_targets, precision, n_memory, step )
%KALMAN FILTERING This function computes an estimation on actual position of
%the target using a Kalman filter. 
    %INPUT:
        % kalman_targets    Target tracks using kalman filter in previous steps
        % precision         Pixels per meter
        % n_memory          Number of previous instants to take into account for the estimation 
        % step              Current step   
    % OUTPUT:
        % kalman targets    Target tracks using kalman filter including current step  
    % NOTATION:
        % X_est(k) --> Current estimation
        % K(k) --> Kalman Gain
        % X_est(k-1) --> Previous estimation
        % Z(k) --> Measured value

        
    % Initializing variables
    A = [1 0; 0 1];  % Prediction weight of previous estimation --> X_est(k) = A*X_est(k-1)
    H = [1 0; 0 1];  % Measurement weigth of current estimation --> Z(k) = H*X_est(k)
    Q = [0.01 0; 0 0.01];  % Process noise covariance matrix 
    R = [0.01 0; 0 0.01];  % Observation noise covariance matrix
    f = 0;

    % Searching memory...
    empty_steps = [];   % Searchs if a previous target exists
    for s = 1:step-1
        for p = 1:length(kalman_targets(:,step))
        empty_steps(p,s) = isstruct(kalman_targets{p,s});
        end
    end
    
    if step>1
        if length(empty_steps(:,1))>1
            empty_steps = sum(empty_steps);
        end
    end
    
    if step <= n_memory  % In case that first target appears in an instant less than N steps
        n = step-1;
    else
        n = n_memory;
    end

    ntargets = length(kalman_targets(:,step));

    for t = 1:ntargets
        if isstruct(kalman_targets{t,step})
            if sum(empty_steps((step-n):(step-1)))>0
                % Target EXISTS in last N steps (frames)
                pos = find((empty_steps(1:(step-1)))>0); % Previous steps with existing targets
                pos = pos((pos<step)&(pos>(step-n)));
                for i = length(pos):-1:1
                    ntargets_pre = length(kalman_targets(:,pos(i)));
                    if ntargets_pre>0
                        for t_pre = 1:ntargets_pre
                            if isstruct(kalman_targets{t_pre,pos(i)})
                                if kalman_targets{t,step}.ID == kalman_targets{t_pre,pos(i)}.ID
                                    if kalman_targets{t,step}.position == kalman_targets{t_pre,pos(i)}.position
                                        kalman_targets{t,step}.X_est = kalman_targets{t_pre,pos(i)}.X_est;
                                        kalman_targets{t,step}.P = kalman_targets{t_pre,pos(i)}.P;
                                        f = 1;
                                        break;
                                    else  
                                        % Kalman filtering

                                        % TIME UPDATE --> PREDICTION
                                        X_est_pre = A*kalman_targets{t_pre,pos(i)}.X_est;  % Prediction 
                                        P_pre = A*kalman_targets{t_pre,pos(i)}.P*A' + Q;   % Error covariance prediction

                                        % MEASUREMENT UPDATE --> CORRECTION
                                        K = P_pre*H'*(H*P_pre*H'+R)^-1; % Kalman gain
                                        kalman_targets{t,step}.X_est = X_est_pre + K*(kalman_targets{t,step}.position'-H*X_est_pre);  % Estimation
                                        kalman_targets{t,step}.P = (eye(2)-K*H)*P_pre; % Error covariance estimation
                                        f = 1;
                                        break;
                                    end
                                end
                            end
                        end
                        if f
                            break;
                        end
                    end
                end
            end
        end
    end
end

