% Executable
clc, clear all, close all
tic
profile on

% Initialization
users = 10;
dimensions = [-12 12; -12 12]; % [x_min x_max; y_min y_max]
total_steps = 200;
calibration_steps = 10;
precision = 0.5;
mean_distance = 7;  % Mean distance in positions
step_distance = 0.8;   % in meters

%% Input
    % Paths that the targets follow inside the considered map
users_path = create_path(users, dimensions, total_steps, step_distance, mean_distance);
users_voxels = floor(1/precision.*(users_path - dimensions(:,1)))+1;    % Voxels in which the user is positioned

user_objects = cell(1,total_steps); % Creating Objects IDs for Performance analysis 

for user = 1:users
    for s = 1:total_steps
        if ~isnan(users_voxels(:,s,user))
            user_objects{user,s} = struct('position',users_voxels(:,s,user),'ID',user);
        end
    end
end

% Plotting users paths
figure('name','Users paths')
for i = 1:users
    plot(users_path(1,:,i),users_path(2,:,i))
    hold on
end
hold on
grid on
title('Paths of targets in the area')
xlabel('X')
ylabel('Y')
axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
%% Generalized gaussian radiation
    % Footprint on the map caused by the presence of a target

% Gaussian radiation characteristics
radiation_amplitude = 20;     % Maximum change in signal, when a target is present
target_width = [1 0.5];   % Target width (meters) in dimensions [x y]
rotation = 0;   % Target rotation in radians
noiselevel = 0.25;  % Noise deviation for the gaussian function (i.e. noiselevel*randn)

% RF radiation
radiation = create_radiation(dimensions, users_path, precision,calibration_steps, radiation_amplitude, target_width, rotation, noiselevel );
%% RF radiation 
    % Measurements based on the change in RSS (received signal strength) in
    % each link (i.e. conexion between two sensors) caused by each voxel
    
% RF model - Change in RSS by each voxel - Simulation
rss_change = radiation;

%% RF sensor links - Change in RSS by each link - Simulation
sensor_position = [-12 -9 -6 0 6 9 12 -12 -6 0 6 12 -12 12 -12 -6 0 6 12 -12 12 -12 -6 0 6 12 -12 -9 -6 0 6 9 12; -12 -12 -12 -12 -12 -12 -12 -6 -6 -6 -6 -6 -9 -9 0 0 0 0 0 9 9 6 6 6 6 6 12 12 12 12 12 12 12];
nsensors = length(sensor_position(1,:));
nlinks = (nsensors^2-nsensors)/2;   % Number of links: L = (K^2-K)/2 (K: number of sensors)
lambda = 5;  % Ellipse excess path length in meters [m]
link_weights = locate_link_ellipses(dimensions,sensor_position,precision,nlinks,lambda);

% Plotting ellipses
figure(2)
k_x = length(dimensions(1,1):precision:dimensions(1,2));
g_x = (dimensions(1,2)-dimensions(1,1))/precision;
k_y = length(dimensions(2,1):precision:dimensions(2,2));
g_y = (dimensions(2,2)-dimensions(2,1))/precision;

x = dimensions(1,1):precision*k_x/g_x:dimensions(1,2);
y = dimensions(2,1):precision*k_y/g_y:dimensions(2,2);

z = zeros(length(y),length(x));
h = sum(link_weights)./length(link_weights(:,1));

for j = 1:length(y)
    for i = 1:length(x)
        z(i,j) = h(length(x)*(j-1)+i);
    end
end

surf(x,y,z)
view(3)
alpha(0.5)
hold on
plot(sensor_position(1,:),sensor_position(2,:),'o','MarkerFaceColor',[92 212 170]./255,'MarkerEdgeColor',[19 53 59]./255)
axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
   
rss_change_link = rss_links(rss_change,link_weights,nlinks); % Received signal (in terms of change of received signal strength) for each link
% For real environment with real sensor measurements of signal strength for
% each link, this value is calculated as
% 1/sum(F_l_c)*sum(F_l_c*diff(Rss-avg_Rss)), where F_l_c is the fade level
% for link l in channel c (if sensors can receive in different frequency
% channels), Rss is the received signal strength at that instant and
% avg_Rss is the average RSS measure at the begining (calibration scenario)

%% Sensors sending quantized decisions
Nbits = 12;
sensors_data = sensor_quantization(rss_change_link, Nbits);

%% Channel BER

channel_ber = 1e-12;
% Resizing sensors_data matrix
line_sensors_data = [];
for l = 1:length(sensors_data(:,1))
    line_sensors_data = [line_sensors_data sensors_data(l,:)];
end

% Converting decimals to binary
bits_sensors_data = zeros(1,Nbits*length(line_sensors_data));
for b = 1:length(line_sensors_data)
    current_bits = de2bi(line_sensors_data(b),'left-msb');
    if current_bits < Nbits
        nbits_zeros = zeros(1,Nbits-length(current_bits));
        current_bits = [nbits_zeros current_bits];
    end
    bits_sensors_data(1,(Nbits*(b-1)+1):(Nbits*b)) = current_bits;
end

% Channel errors
nbits_error = channel_ber*length(bits_sensors_data);
rand_position = randi([1 length(bits_sensors_data)], 1, floor(nbits_error));

% Data with errors
bits_sensors_data(rand_position) = 1 - bits_sensors_data(rand_position);

% Reconstructing values in a row matrix
reconstructed_line_sensors_data = zeros(size(line_sensors_data));
for e = 1:length(line_sensors_data)
    reconstructed_line_sensors_data(e) = bi2de(bits_sensors_data((Nbits*(e-1)+1):(e*Nbits)),'left-msb');
end

% Resizing data sensors matrix
new_sensors_data = zeros(size(sensors_data));
for g = 1:length(sensors_data(:,1))
    new_sensors_data(g,:) = reconstructed_line_sensors_data(1,(length(sensors_data(1,:))*(g-1)+1):(length(sensors_data(1,:))*g));
end

length(find(sensors_data ~= new_sensors_data))

sensors_data = new_sensors_data;
%% RTI Image Estimation
% This section is part of the estimation algorithm, after the simulation
% (or real environment test)
% Using Regularized Least-Squares (RLS) approach, the RSS change in each
% voxel is estimated (in simulation this is called "rss_change")

% The Fusion Center (FC) estimates the radiation from the sensors measurements

% Noise deviation - Regularization parameter
sigma_n = 1;
% Voxel measurements' deviation
sigma_x_dB = 0.2236; %dB
% Voxels' correlation distance
delta_c = 3;
% RSS estimation 
rss_change_estimate = rss_estimation(dimensions, sensors_data, link_weights, precision, sigma_n, sigma_x_dB, delta_c);

%% Multiple Target Tracking with RTI images (RF sensors)
global istarget;
istarget = false;   % Empty area or monitoring targets
% Initialization of variables
filtered_rti = zeros(size(rss_change_estimate(:,:,calibration_steps+1:end)));
alfa_f = 0.9;
beta = 0.8;
xaxis = (dimensions(1,1)+precision/2):precision:(dimensions(1,2)-precision/2);
yaxis = (dimensions(2,1)+precision/2):precision:(dimensions(2,2)-precision/2);
total_clusters = cell(1,total_steps);

% Threshold when the monitored area is empty
Ie = mean(max(max(rss_change_estimate(:,:,1:calibration_steps)))); % Average maximum intensity in the training period
Tt = [2*Ie 0];  % Initial Detection Threshold
% Initizalization of Threshold when targets are being tracked
I_lpf = ones(1,total_steps);

% Tracking algorithm initialization
% Parameters
Re_width = 2; % Entrance/Exit region width in meters
radius_t = 2.5;   % Initial radius around a target in meters [m]
n_memory = 5;   % Number of previous decisions (clusters) taken into account for instant k
n_confirm = 5;  % To confirm a track, it has to be assigned to the same ID at least 'n_confirm' times in the last 'n_max_confirm' observations
n_max_confirm = 10;
n_delete = 5;   % To delete a track, it has to be not assigned in the last 'n_delete' observations

% Initializing tracking variables:
targets = cell(1,total_steps);
kalman_targets = cell(1,total_steps);
particle_targets = cell(1,total_steps);
target_id = 1;
confirmed_tracks = [];  % Confirmed tracks after N of M observations
track_candidates = [];  % Track candidates to be confirmed or deleted
deleted_tracks = [];    % Deleted track candidates if not confirmed

% Performance analysis - Initializing variables
data_association = zeros(users,5,total_steps);  % Object-Hypothesis
total_deviation = 0;
total_deviation_kalman = 0;
total_deviation_particle = 0;
n_matches = 0;
total_false_positives = 0;
total_misses = 0;
total_mismatches = 0;

% Clustering parameters
Tc_inmeters = 2;  % Threshold for clustering (aprox. for a human width) in meters [m]

% Particle filter Initialization
ids = [];
particle_filters = [];
particle_boundaries_x = floor(1/precision.*([dimensions(1,1);dimensions(1,2)-precision] - dimensions(1,1)))+1;    % Particle boundaries in X-axis
particle_boundaries_y = floor(1/precision.*([dimensions(2,1);dimensions(2,2)-precision] - dimensions(2,1)))+1;    % Particle boundaries in Y-axis
particle_boundaries = [particle_boundaries_x' ; particle_boundaries_y'];

figure(3)
hold on
subplot(3,3,1)  % Subplot 1 -- Users Paths
for i = 1:users
    plot(users_path(1,:,i),users_path(2,:,i))
    hold on
end
grid on
title('Users path')
xlabel('X')
ylabel('Y')
axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])

for step = 1:total_steps

    % Thresholding
    filtered_rti(:,:,step) = detection_thresholding(rss_change_estimate(:,:,calibration_steps+step),Tt);    
    % Clustering  
    [clusters_rti, total_clusters{1,step}, voxels_forcluster] = clustering(filtered_rti(:,:,step), precision, Tc_inmeters);
    % Tracking
    [targets, track_candidates, confirmed_tracks, deleted_tracks, target_id ] = target_tracking( targets, total_clusters, step,...
    dimensions, precision, Re_width, radius_t, n_memory, n_confirm, n_max_confirm, n_delete, target_id, track_candidates, ...
    confirmed_tracks, deleted_tracks );
 
    % Kalman filtering
    for t = 1:length(targets(:,step))
        kalman_targets{t,step} = targets{t,step};
    end
    kalman_targets = kalman_filtering( kalman_targets, precision, n_memory, step );
    
    % Particle filtering
    for t = 1:length(targets(:,step))
        particle_targets{t,step} = targets{t,step};
    end
    [ particle_targets, particle_filters, ids] = particle_filtering( particle_targets, particle_filters, particle_boundaries, ids, n_memory, step );
    
    % Object-Hypothesis association
    data_association = object_hypothesis_association(user_objects, targets, kalman_targets, particle_targets, step, data_association);
    
    % Performance analysis
    [ total_deviation, total_deviation_kalman, total_deviation_particle, n_matches, MOTP, MOTP_kalman, MOTP_particle, max_deviation, max_deviation_kalman, max_deviation_particle ] = tracking_deviation( data_association, step, total_deviation, total_deviation_kalman, total_deviation_particle, n_matches, precision );
    [ total_false_positives ] = error_false_positives( targets, data_association, step, total_false_positives );
    [ total_misses ] = error_misses( data_association, total_misses, step );
    [ total_mismatches ] = error_mismatches( data_association, total_mismatches, step );
    
    if istarget     
        % Iterative thresholding
        Imin_previous = zeros(1,length(voxels_forcluster(:,1)));
        for pos = 1:length(voxels_forcluster(:,1))
            Imin_previous(pos) = filtered_rti(voxels_forcluster(pos,1),voxels_forcluster(pos,2),step); 
        end   
        Imin = min(Imin_previous);
        I_lpf(step+1) = alfa_f*I_lpf(step)+(1-alfa_f)*Imin; % Low-pass filter of Imin
        Tt(2) = beta*I_lpf(step+1); 
        
        %% Plotting  
        subplot(3,3,2)
        cla
        subplot(3,3,3)
        cla
        subplot(3,3,4)
        cla
        subplot(3,3,5)
        cla
        subplot(3,3,9)
        cla
        
        rows_inmeters = (voxels_forcluster(:,1).*precision+dimensions(1,1))-precision/2;
        columns_inmeters = (voxels_forcluster(:,2).*precision+dimensions(2,1))-precision/2;
        
        subplot(3,3,2)  % Subplot 2 -- Input radiation
        surf(xaxis,yaxis,rss_change(:,:,calibration_steps+step)','EdgeColor','none')
        view(2);
        colormap jet;
        grid on;
        caxis([0 max(max(rss_change(:,:,calibration_steps+step)))]);
        grid on
        title('Input Radiation')
        xlabel('X')
        ylabel('Y')
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])

        subplot(3,3,3)  % Subplot 3 -- Estimated radiation from sensors
        surf(xaxis,yaxis,rss_change_estimate(:,:,calibration_steps+step)','EdgeColor','none')
        view(2);
        colormap jet;
        grid on;
        caxis([0 max(max(rss_change_estimate(:,:,calibration_steps+step)))]);
        grid on
        title('Estimated Radiation')
        xlabel('X')
        ylabel('Y')
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        
        subplot(3,3,4) % Subplot 4 -- Filtered estimated radiation 
        surf(xaxis,yaxis,filtered_rti(:,:,step)','EdgeColor','none')
        view(2)
        caxis([0 max(max(filtered_rti(:,:,step)))]);
        hold on
        t = linspace(0,2*pi);
        for w = 1:length(targets(:,step))
            if isstruct(targets{w,step})
                radius_inmeters = (targets{w,step}.radius).*precision;
                position_x_inmeters = ((targets{w,step}.position(1)).*precision+dimensions(1,1))-precision/2;
                position_y_inmeters = ((targets{w,step}.position(2)).*precision+dimensions(2,1))-precision/2;
                plot(position_x_inmeters+radius_inmeters.*cos(t),position_y_inmeters+radius_inmeters.*sin(t))
                hold on
            end
        end
        grid on;
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Filtered estimated RTI')
        
        subplot(3,3,5) % Subplot 5 -- Clusters
        scatter(rows_inmeters,columns_inmeters,10,clusters_rti)
        hold on
        t = linspace(0,2*pi);
        for w = 1:length(targets(:,step))
            if isstruct(targets{w,step})
                radius_inmeters = (targets{w,step}.radius).*precision;
                position_x_inmeters = ((targets{w,step}.position(1)).*precision+dimensions(1,1))-precision/2;
                position_y_inmeters = ((targets{w,step}.position(2)).*precision+dimensions(2,1))-precision/2;
                plot(position_x_inmeters+radius_inmeters.*cos(t),position_y_inmeters+radius_inmeters.*sin(t))
                hold on
            end
        end
        grid on;
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Clusters in estimated RTI')
        
        subplot(3,3,6)  % Subplot 6 -- Multi target Tracking
        hold on
        for b = 1:length(targets(:,step))
            if isstruct(targets{b,step})
                subplot(3,3,6)
                target_x_position_inmeters = ((targets{b,step}.position(1)).*precision+dimensions(1,1))-precision/2;
                target_y_position_inmeters = ((targets{b,step}.position(2)).*precision+dimensions(2,1))-precision/2;
                plot(target_x_position_inmeters,target_y_position_inmeters,'w*')
                hold on
                ln = findobj(subplot(3,3,6),'type','line');
                set(ln, 'marker', 'x', 'markers', 1, 'markerfa', 'w')
                if targets{b,step}.target_type == 'candidate'
                    text(target_x_position_inmeters,target_y_position_inmeters,num2str(targets{b,step}.ID), 'Color', [255 152 67]./255)
                elseif targets{b,step}.target_type == 'confirmed'
                    text(target_x_position_inmeters,target_y_position_inmeters,num2str(targets{b,step}.ID), 'Color', [57 232 74]./255)
                else
                    text(target_x_position_inmeters,target_y_position_inmeters,num2str(targets{b,step}.ID), 'Color', 'k')
                end
                grid on;
                axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
                xlabel('[m]');
                ylabel('[m]');
                title('Multi Target Tracking')
                subplot(3,3,1)             
                plot(target_x_position_inmeters,target_y_position_inmeters,'x','Color','r')
                hold on
                
            end
        end
        
        subplot(3,3,7)  % Subplot 7 -- Kalman Filtering 
        hold on
        for b = 1:length(kalman_targets(:,step))
            if isstruct(kalman_targets{b,step}) 
                
                    subplot(3,3,7)
                    target_x_position_inmeters = ((kalman_targets{b,step}.X_est(1)).*precision+dimensions(1,1))-precision/2;
                    target_y_position_inmeters = ((kalman_targets{b,step}.X_est(2)).*precision+dimensions(2,1))-precision/2;
                    plot(target_x_position_inmeters,target_y_position_inmeters,'x','Color','b')
                    grid on;
                    axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
                    xlabel('[m]');
                    ylabel('[m]');
                    title('Kalman Filtering')
                    subplot(3,3,1)             
                    plot(target_x_position_inmeters,target_y_position_inmeters,'o','Color','b')
                    hold on
            end
        end
        subplot(3,3,8)  % Subplot 8 -- Particle Filtering
        hold on
        for b = 1:length(particle_targets(:,step))
            if isstruct(particle_targets{b,step}) 
                
                    subplot(3,3,8) 
                    target_x_position_inmeters = ((particle_targets{b,step}.particle_X_est(1)).*precision+dimensions(1,1))-precision/2;
                    target_y_position_inmeters = ((particle_targets{b,step}.particle_X_est(2)).*precision+dimensions(2,1))-precision/2;
                    plot(target_x_position_inmeters,target_y_position_inmeters,'x', 'Color', 'm')
                    grid on;
                    axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
                    xlabel('[m]');
                    ylabel('[m]');
                    title('Particle Filtering')
                    subplot(3,3,1)             
                    plot(target_x_position_inmeters,target_y_position_inmeters,'s','Color','g')
                    hold on
                
            end
        end 
        subplot(3,3,9) % Subplot 9 -- Observations vs. targets
        if step>1
            for c = 1:length(targets(:,step-1))
                if isstruct(targets{c,step-1})
                    target_x_position_inmeters = ((targets{c,step-1}.position(1)).*precision+dimensions(1,1))-precision/2;
                    target_y_position_inmeters = ((targets{c,step-1}.position(2)).*precision+dimensions(1,1))-precision/2;
                    plot(target_x_position_inmeters,target_y_position_inmeters,'x')
                    grid on;
                    axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
                    xlabel('[m]');
                    ylabel('[m]');
                    title('Observations vs. targets')
                    %legend('o -> observations in k','x -> targets in k-1')
                    hold on
                end
            end
            for d = 1:length(targets(:,step))
                if isstruct(targets{d,step})
                    observation_x_position_inmeters = ((targets{d,step}.position(1)).*precision+dimensions(1,1))-precision/2;
                    observation_y_position_inmeters = ((targets{d,step}.position(2)).*precision+dimensions(1,1))-precision/2;
                    plot(observation_x_position_inmeters,observation_y_position_inmeters,'o')
                    hold on
                    radius_inmeters = (targets{d,step}.radius).*precision;
                    plot(observation_x_position_inmeters+radius_inmeters.*cos(t),observation_y_position_inmeters+radius_inmeters.*sin(t))
                    grid on;
                    axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
                    xlabel('[m]');
                    ylabel('[m]');
                    title('Observations vs. targets')
                    
                    hold on
                    
                end
            end
            
        end  
         pause(0.2)
           
    else
        subplot(3,3,2)
        cla
        subplot(3,3,3)
        cla
        subplot(3,3,4)
        cla
        subplot(3,3,5)
        cla
        subplot(3,3,8)
        
        subplot(3,3,2)  % Subplot 2 -- Input radiation
        surf(xaxis,yaxis,rss_change(:,:,calibration_steps+step)','EdgeColor','none')
        view(2);
        colormap jet;
        grid on;
        caxis([0 max(max(rss_change(:,:,calibration_steps+step)))]);
        grid on
        title('Input Radiation')
        xlabel('X')
        ylabel('Y')
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])

        subplot(3,3,3)  % Subplot 3 -- Estimated radiation from sensors
        surf(xaxis,yaxis,rss_change_estimate(:,:,calibration_steps+step)','EdgeColor','none')
        view(2);
        colormap jet;
        grid on;
        caxis([0 max(max(rss_change_estimate(:,:,calibration_steps+step)))]);
        grid on
        title('Estimated Radiation')
        xlabel('X')
        ylabel('Y')
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        
        subplot(3,3,4) % Subplot 4 -- Filtered radiation
        surf(xaxis,yaxis,filtered_rti(:,:,step)','EdgeColor','none')
        view(2)
        caxis([0 radiation_amplitude]);
        grid on;
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Filtered estimated RTI')
        
        subplot(3,3,5) % Subplot 5 -- Clusters
        plot(0,0);
        grid on;
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Clusters in estimated RTI')
       
                
        subplot(3,3,6) % Subplot 6 -- Multi target tracking
        plot(0,0);
        hold on
        grid on;
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Multi Target Tracking')
        
        
        subplot(3,3,7) % Subplot 7 -- Kalman Filtering
        plot(0,0);
        hold on
        grid on;
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Kalman Filtering')
        
        subplot(3,3,8) % Subplot 8 -- Particle Filtering
        plot(0,0);
        hold on
        grid on 
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Particle Filtering')
        
        subplot(3,3,9) % Subplot 9 -- Observations vs. targets
        plot(0,0);
        hold on
        grid on 
        axis([dimensions(1,1) dimensions(1,2) dimensions(2,1) dimensions(2,2)])
        xlabel('[m]');
        ylabel('[m]');
        title('Observations vs. targets')
        pause(0.1)
    end    
end

profile viewer
toc