function [ data_association ] = object_hypothesis_association( user_objects, targets, kalman_targets, particle_targets, step, data_association )
%OBJECT_HYPOTHESIS_ASSOCIATION This function associates one input position
%of a real target to an hypothesis of an estimated possition after the
%tracking. This is made according to the input objects IDs and the output
%hypotheses IDs in each step (instant). This function is important for comparing and evaluating
%the errors and to analyze the performance of the tracker. 

%   INPUT
%       user_objects        Array (n_users x n_steps) with the input Objects IDs and positions for each step
%       targets             Array (n_hypothesis x n_steps) with the ouput Hypotheses IDs and position for each step, and other track characteristics (ID, position, radius, target_type)
%       kalman_targets      Array (n_hypothesis x n_steps) with the ouput Hypotheses IDs and position for each step after Kalman filtering (ID, position, radius, target_type, X_est, P)
%       step                Current instant
%       data_association    Object-Hypothesis mapping in previous steps
%   OUTPUT
%       data_association    Array ((2 IDs + dmin + dmin_kalman) x (n_object2hypothesis_associations) x n_steps) in current step

    nusers = length(user_objects(:,step));
    ntargets = length(targets(:,step));
    
    for u = 1:nusers
        if isstruct(user_objects{u,step})
            data_association(u,1,step) = [user_objects{u,step}.ID];
        end
    end
    for t = 1:ntargets
        if isstruct(targets{t,step})
            t_distances = [];
            for u = 1:nusers
                if isstruct(user_objects{u,step})
                    t_distances = [t_distances norm(targets{t,step}.position'-user_objects{u,step}.position)];
                else
                    t_distances = [t_distances Inf];
                end
            end
            
            % Find minimum distance between user objects and target assignations in current step
            i_user = find(t_distances<targets{t,step}.radius);
            
            if ~isempty(i_user) % There is a match between real data and estimation
                if length(i_user)>1
                    a = 1;
                end
                if sum(data_association(:,2,step)) == 0
                    % FIRST ASSOCIATION IN STEP
                    [d_min, min_position] = min(t_distances);
                    data_association(min_position,2,step) = [targets{t,step}.ID];
                    data_association(min_position,3,step) = d_min;
                    if sum(kalman_targets{t,step}.X_est)>0
                        data_association(min_position,4,step) = norm(kalman_targets{t,step}.X_est-user_objects{min_position,step}.position);
                    end
                    if sum(particle_targets{t,step}.particle_X_est)>0
                        data_association(min_position,5,step) = norm(particle_targets{t,step}.particle_X_est'-user_objects{min_position,step}.position);
                    end
                else
                    j_user = find((data_association(:,1,step)>0) & (data_association(:,2,step) == 0));
                    if ~isempty(j_user)
                        removing_assigned_targets = Inf(1,nusers);
                        removing_assigned_targets(j_user) = 1;
                        for r = 1:length(removing_assigned_targets)
                            if removing_assigned_targets(r) == Inf
                                t_distances(r) = Inf;
                            end
                        end
                        % ADDING NEW ASSOCIATION IN STEP
                        [d_min, min_position] = min(t_distances);
                        data_association(min_position,2,step) = [targets{t,step}.ID];
                        data_association(min_position,3,step) = d_min;
                        if sum(kalman_targets{t,step}.X_est)>0
                            data_association(min_position,4,step) = norm(kalman_targets{t,step}.X_est-user_objects{min_position,step}.position);
                        end
                        if sum(particle_targets{t,step}.particle_X_est)>0
                            data_association(min_position,5,step) = norm(particle_targets{t,step}.particle_X_est'-user_objects{min_position,step}.position);
                        end
                    end
                end
            end     
        end
    end  
end